from pact_test.either import *
from pact_test.matchers.matcher import *


def match(actual, expected):
    return _match_path(actual, expected)\
        .concat(_match_query, expected)\
        .concat(_match_method, expected)\
        .concat(_match_headers, expected)\
        .concat(_match_body, expected)


def _match_body(actual, expected):
    expected_body = expected.body
    actual_body = actual.body

    if expected_body is None and actual_body is None:
        return Right(actual)

    if is_string(expected_body) and is_string(actual_body) and expected_body == actual_body:
        return Right(actual)

    if type(expected_body) is dict \
            and type(actual_body) is dict \
            and match_dicts_all_keys_and_values(expected_body, actual_body):
            return Right(actual)

    return Left(build_error_message('body', expected_body, actual_body))


def _match_headers(actual, expected):
    actual_dict = dict(pair for d in actual.headers for pair in d.items())
    expected_dict = dict(pair for d in expected.headers for pair in d.items())

    insensitive_actual = {k.upper(): v for (k, v) in actual_dict.items()}
    insensitive_expected = {k.upper(): v for (k, v) in expected_dict.items()}

    if is_subset(insensitive_expected, insensitive_actual):
        return Right(actual)
    return Left(build_error_message('headers', expected.headers, actual.headers))


def _match_path(actual, expected):
    if actual.path == expected.path:
        return Right(actual)
    return Left(build_error_message('path', expected.path, actual.path))


def _match_query(actual, expected):
    if actual.query == expected.query:
        return Right(actual)
    return Left(build_error_message('query', expected.query, actual.query))


def _match_method(actual, expected):
    if actual.method.upper() == expected.method.upper():
        return Right(actual)
    return Left(build_error_message('method', expected.method.upper(), actual.method.upper()))



import os
import imp
import inspect
from pact_test.either import *
from pact_test import PactHelper
from pact_test.constants import MISSING_SETUP
from pact_test.constants import MISSING_TEAR_DOWN
from pact_test.constants import EXTEND_PACT_HELPER
from pact_test.constants import MISSING_PACT_HELPER


def load_pact_helper(consumer_tests_path):
    return _path_to_pact_helper(consumer_tests_path)\
               .concat(_load_module, 'pact_helper') \
               >> _load_user_class


def _load_user_class(user_module):
    user_class = None

    for name, obj in inspect.getmembers(user_module):
        if inspect.isclass(obj) and len(inspect.getmro(obj)) > 2 and issubclass(obj, PactHelper):
            user_class = obj()

    if user_class is None:
        return Left(EXTEND_PACT_HELPER)
    if hasattr(user_class, 'setup') is False:
        return Left(MISSING_SETUP)
    if hasattr(user_class, 'tear_down') is False:
        return Left(MISSING_TEAR_DOWN)

    return Right(user_class)


def _load_module(path, module_name):
    try:
        return Right(imp.load_source(module_name, path))
    except Exception:
        return Left(MISSING_PACT_HELPER + path + '".')


def _path_to_pact_helper(consumer_tests_path):
    path = os.path.join(consumer_tests_path, 'pact_helper.py')
    if os.path.isfile(path) is False:
        msg = MISSING_PACT_HELPER + consumer_tests_path + '".'
        return Left(msg)
    return Right(path)



from pact_test.either import *

PREFIX = '[Pact Test for Python] - '                        # pragma: no cover


def info(message):                                          # pragma: no cover
    print('\033[92m' + PREFIX + str(message) + '\033[0m')   # pragma: no cover


def error(message):                                         # pragma: no cover
    print('\033[91m' + PREFIX + str(message) + '\033[0m')   # pragma: no cover


def debug(message):                                         # pragma: no cover
    print('\033[93m' + PREFIX + str(message) + '\033[0m')   # pragma: no cover


def log_consumers_test_results(test_results):
    if type(test_results) is Left:
        error(test_results.value)
    else:
        if type(test_results.value) is Left:
            error(test_results.value.value)
        else:
            for test_result in test_results.value:
                print()
                info('Test: ' + test_result.value['test'])
                for result in test_result.value['results']:
                    info('  GIVEN ' + result.value['state'] + ' UPON RECEIVING ' + result.value['description'])
                    info('    status: ' + result.value['status'])
                    for test_error in result.value['errors']:
                        error('      expected: ' + str(test_error['expected']))
                        error('      actual:   ' + str(test_error['actual']))
                        error('      message:  ' + str(test_error['message']))
    info('')
    info('Goodbye!')
    print()


def log_providers_test_results(test_results):
    if type(test_results) is Left:
        error(test_results.value)
    else:
        for r in test_results.value:
            print()
            info('A pact between ' + r['consumer']['name'] + ' and ' + r['provider']['name'])
            for i in r['interactions']:
                debug(i)
                info('')
                info('  Given ' + i['providerState'] + ', upon receiving ' + i['description'] + ' from ' + r['consumer']['name'] + ' with:')
                info('')
                info('  {')
                info('    "method": ' + str(i['request']['method']) + ',')
                info('    "path": ' + str(i['request']['path']) + ',')
                info('    "query": ' + str(i['request']['query']) + ',')
                info('    "headers": ' + str(i['request']['headers']) + ',')
                info('    "body": ' + str(i['request']['body']))
                info('  }')
                info('')
                info('  ' + r['provider']['name'] + ' will respond with: ')
                info('')
                info('  {')
                info('    "status": ' + str(i['response']['status']) + ',')
                info('    "body": ' + str(i['response']['body']) + ',')
                info('    "headers": ' + str(i['response']['headers']))
                info('  }')
                info('')



import json
import requests
from pact_test.either import *


PACT_BROKER_URL = 'http://localhost:9292/'


def upload_pact(provider_name, consumer_name, pact, base_url=PACT_BROKER_URL):
    print()
    pact = format_headers(pact)
    current_version = get_latest_version(consumer_name)
    if type(current_version) is Right:
        v = next_version(current_version.value)
        try:
            url = base_url + 'pacts/provider/' + provider_name + '/consumer/' + consumer_name + '/version/' + v
            payload = json.dumps(pact)
            headers = {'content-type': 'application/json'}
            response = requests.put(url, data=payload, headers=headers)
            return Right(response.json())
        except requests.exceptions.ConnectionError as e:
            msg = 'Failed to establish a new connection with ' + base_url
            return Left(msg)
    print()


def get_latest_version(consumer_name, base_url=PACT_BROKER_URL):
    try:
        url = base_url + 'pacticipants/' + consumer_name + '/versions/'
        response = requests.get(url)
        if response.status_code is not 200:
            return Right('1.0.0')
        return Right(response.json()['_embedded']['versions'][0]['number'])
    except requests.exceptions.ConnectionError as e:
        msg = 'Failed to establish a new connection with ' + base_url
        return Left(msg)


def next_version(current_version='1.0.0'):
    versions = current_version.split('.')
    next_minor = str(1 + int(versions[-1]))
    return '.'.join([versions[0], versions[1], next_minor])


def format_headers(pact):
    for interaction in pact.get('interactions', []):
        req_headers = interaction.get('request').get('headers')
        fixed_req_headers = {}
        for h in req_headers:
            fixed_req_headers.update(h)
        interaction['request']['headers'] = fixed_req_headers

        res_headers = interaction.get('response').get('headers')
        fixes_req_headers = {}
        for h in res_headers:
            fixes_req_headers.update(h)
        interaction['response']['headers'] = fixes_req_headers
    return pact



class PactRequest(object):
    def __init__(self, method='', body, headers=[], path='', query=''):
        print('BODY??? >>> ' + str(body))
        self.body = body
        print('BODY??? >>> ' + str(self.body))
        self.path = path
        self.query = query
        self.method = method
        self.headers = headers

    def __str__(self):
        return "$$$ BODY: " + str(self.body) + ", HEADERS: " + str(self.headers)



import os
import imp
import inspect
from pact_test.either import *
from pact_test.constants import *
from pact_test.utils.logger import error
from pact_test.utils.logger import debug
from pact_test.runners.service_providers.request_test import verify_request


class ServiceProviderTestSuiteRunner(object):
    def __init__(self, config):
        self.config = config

    def verify(self):
        debug('Verify providers: START')
        tests = self.collect_tests()
        if type(tests) is Right:
            debug(str(len(tests.value)) + ' test(s) found.')
            pacts = []
            for test in tests.value:
                test_verification = test.is_valid()
                if type(test_verification) is Right:
                    pact = self.create_pact(test)
                    if len(pact['interactions']) == 0:
                        error('Verify providers: EXIT WITH ERRORS:')
                        return Left('No pact-test methods available in test class')
                    pacts.append(pact)
                else:
                    error('Verify providers: EXIT WITH ERRORS:')
                    error(test_verification.value)
                    return test_verification
            return Right(pacts)
        error('Verify providers: EXIT WITH ERRORS:')
        error(tests.value)
        return tests

    @staticmethod
    def create_pact(test):
        interactions = []
        for decorated_method in test.decorated_methods:
            interactions.append(verify_request(decorated_method).value)
        pact = {
            'interactions': interactions,
            'provider': {
                'name': test.has_pact_with
            },
            'consumer': {
                'name': test.service_consumer
            }
        }
        return pact

    def collect_tests(self):
        root = self.config.provider_tests_path
        debug('Root for Provider Tests: ' + str(root))
        all_files = self.all_files()
        if type(all_files) is Right:
            files = list(filter(self.filter_rule, all_files.value))
            files = list(map(lambda f: os.path.join(root, f), files))
            tests = []
            for idx, filename in enumerate(files):
                test = imp.load_source('test' + str(idx), filename)
                for name, obj in inspect.getmembers(test):
                    if inspect.isclass(obj) and len(inspect.getmro(obj)) > 2:
                        test_parent = inspect.getmro(obj)[1].__name__
                        if test_parent == PROVIDER_TEST_PARENT:
                            tests.append(obj())

            if not files:
                return Left(MISSING_PROVIDER_TESTS)
            return Right(tests)
        return all_files

    @staticmethod
    def filter_rule(filename):
        return filename != '__init__.py' and filename.endswith('.py')

    def all_files(self):
        try:
            return Right(os.listdir(self.config.provider_tests_path))
        except Exception as e:
            return Left(str(e))
